/* 
 * $Id: bgp_dump.c,v 1.6 1998/05/08 15:02:12 masaki Exp $
 */

#include <fcntl.h>
#include <time.h>

#include <mrt.h>
#include <timer.h>
#include <trace.h>
#include <select.h>
#include <interface.h>
#include <rib.h>
#include <bgp.h>
#ifdef HAVE_IPV6
#include <api6.h>
#endif /* HAVE_IPV6 */


void
bgp_write_mrt_msg (bgp_peer_t *peer, enum MRT_MSG_BGP_TYPES bgptype, 
		   u_char *buf, int length) {
    u_char tmpx[MAX_MSG_SIZE], *cp = tmpx;
    int len = length;
    u_char sync[MAXLINE], *xp = sync;
    int synclen = 0;
    time_t now;
    enum MRT_MSG_TYPES msgtype = MSG_PROTOCOL_BGP;
    int viewno = peer->view;
    struct tm my_tm;

    if (BGP->dump_update_form == NULL && BGP->dump_route_form[viewno] == NULL)
	return;

    pthread_mutex_lock (&BGP->mutex_lock);
    time (&now);

    if (BGP->dump_route_form[viewno]) {

        if (BGP->dump_route_time[viewno] + BGP->dump_route_interval[viewno] 
		<= now) {

	    char name[MAXLINE];
	    struct tm *tm;
	    int fd;
	    uii_connection_t uii;
	    static int nerror = 0;

	    tm = localtime (&now);
            strftime (name, sizeof (name), BGP->dump_route_form[viewno], tm);
	    if ((fd = open (name, (O_CREAT | O_APPEND | O_WRONLY), 0666)) < 0) {
		if (nerror++ <= 0) {
		    trace (TR_ERROR, MRT->trace, 
		           "can not open dump file: %s (%s)\n",
		           name, strerror (errno));
		}
		else if (nerror >= 100) {
		    trace (TR_ERROR, MRT->trace, 
		           "give up opening dump file: %s (%s)\n",
		           name, strerror (errno));
		    Delete (BGP->dump_route_form[viewno]);
    		    BGP->dump_route_form[viewno] = NULL;
		}
		BGP->dump_route_time[viewno] = now;
	    }
	    else {
	        lseek (fd, 0, SEEK_END);
		memset (&uii, 0, sizeof (uii));
	        uii.sockfd = -1;
	        uii.sockfd_out = fd;
		if (bgptype == MSG_BGP_UPDATE) {
	            bgp_dump_view (&uii, BGP->views[viewno], NULL, NULL);
		}
		else {
		    view_open (BGP->views[viewno]);
	            bgp_dump_view (&uii, BGP->views[viewno], NULL, NULL);
		    view_close (BGP->views[viewno]);
		}
	        close (fd);
		BGP->dump_route_time[viewno] = now;

		/* set sync information */
		if (BGP->dump_update_form) {
        	    BGP_PUT_SHORT (viewno, xp);
		    strcpy (xp, name);
		    synclen = sizeof (short) + strlen (name) + 1 /* for zero */;
		}
	    }
	}
    }

    if (BGP->dump_update_form) {
        static io_t *IO = NULL;

        if (IO == NULL) {
	    IO = New_IO (NULL);
        }
  
        if (BGP->dump_update_time == 0 ||
	   (BGP->dump_update_interval && 
		BGP->dump_update_time + BGP->dump_update_interval <= now)) {

	    char name[MAXLINE];
	    struct tm *tm;
	    static int nerror = 0;

#ifdef HAVE_LOCALTIME_R
	    tm = localtime_r (&now, &my_tm);
#else
	    tm = localtime (&now);
#endif /* HAVE_LOCALTIME_R */
            strftime (name, sizeof (name), BGP->dump_update_form, tm);
	    if (io_set (IO, IO_OUTFILE, name, NULL) < 0) {
		if (nerror++ <= 0) {
		    trace (TR_ERROR, MRT->trace, 
		           "can not open dump file: %s (%s)\n",
		           name, strerror (errno));
		}
		else if (nerror >= 100) {
		    trace (TR_ERROR, MRT->trace, 
		           "give up opening dump file: %s (%s)\n",
		           name, strerror (errno));
		    Delete (BGP->dump_update_form);
    		    BGP->dump_update_form = NULL;
		}
		BGP->dump_update_time = now;
    		pthread_mutex_unlock (&BGP->mutex_lock);
		return;
	    }
	    else {
		nerror = 0;
		BGP->dump_update_time = now;
	    }
	}

        /* from */
        BGP_PUT_SHORT (peer->gateway->AS, cp);

#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
            memcpy (cp, prefix_tochar (peer->gateway->prefix), 16);
            cp += 16;
        }
        else
#endif /* HAVE_IPV6 */
        BGP_PUT_NETLONG (prefix_tolong (peer->gateway->prefix), cp);

/*
 * To be compatible with Craig's code, AS and addr for 'to' 
 * should not be included in this MRT message at the moment. 
 *   -- masaki
 */
if (bgptype != MSG_BGP_STATE_CHANGE) {

        /* to */
        BGP_PUT_SHORT (BGP->my_as, cp);
#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
            if (peer->local_addr) {
		assert (peer->local_addr->family == AF_INET6);
                memcpy (cp, prefix_tochar (peer->local_addr), 16);
	    }
	    else {
		memset (cp, 0, 16);
	    }
            cp += 16;
        }
        else
#endif /* HAVE_IPV6 */
	if (peer->local_addr)
            BGP_PUT_NETLONG (prefix_tolong (peer->local_addr), cp);
	else
            BGP_PUT_NETLONG (0L, cp);
}

        memcpy (cp, buf, length);
        len += (cp - tmpx);

#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
	    if (BIT_TEST (peer->options, BGP_BGP4PLUS_01))
		msgtype = MSG_PROTOCOL_BGP4PLUS_01;
	    else
		msgtype = MSG_PROTOCOL_BGP4PLUS;
        }
#endif /* HAVE_IPV6 */

 	/* flush sync info */
	if (synclen > 0)
            io_write (IO, 0, msgtype, MSG_BGP_SYNC, synclen, (char *)sync);

        io_write (IO, 0, msgtype, bgptype, len, (char *)tmpx);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);
}


int
bgp_write_status_change (bgp_peer_t *peer, u_short state) {
    u_char buf[2+2], *cp = buf;

    BGP_PUT_SHORT (peer->state, cp); /* previous */
    BGP_PUT_SHORT (state, cp); /* new */
    bgp_write_mrt_msg (peer, MSG_BGP_STATE_CHANGE, buf, 4);
    return (1);
}

