/*
 * $Id: process.c,v 1.14 1998/05/13 16:19:46 labovit Exp $
 */

#include <stdio.h>
#include <string.h>
#include <mrt.h>
#include <trace.h>
#include <interface.h>
#include <rip.h>
#include <time.h>
#include <bgp.h>
#include <signal.h>
#include <config_file.h>
#include <fcntl.h>
#include <io.h>
#include <dirent.h>
#include "rtracker.h"


extern bgp_t *BGP;
extern rip_t *RIP;
extern trace_t *default_trace;

void rtr_process_bgp_update (mrt_msg_t * msg, irr_connection_t *irr);
void rtr_update_route (rtracker_peer_t *peer, prefix_t *prefix, bgp_attr_t *attr);
void rtr_update_peer (int flag, rtracker_peer_t *tmp, prefix_t *prefix, bgp_attr_t *attr);
int count_routes (rtracker_peer_t *peer);

/* rtr_process_input
 * Run through a given file and process it... 
 */
void rtr_process_input (char *file, uii_connection_t *uii, irr_connection_t *irr) {
  mrt_msg_t *msg;
  io_t *IO;
  
  IO = New_IO (default_trace);

  trace (NORM, default_trace, "Opening %s\n", file);

  if (io_set (IO, IO_INFILE, file, 0)) {
    trace (NORM, default_trace, "Failed to open file  %s\n", file);
    uii_send_data (uii, "Failed to open file %s\r\n", file);
    Delete (IO);
    perror ("Could not open file\n");
    return;
  }

  while (1) {
    if ((msg = (mrt_msg_t *) io_read (IO)) == NULL) {
      /*       printf("\nInvalid MSG "); */
      io_set (IO, IO_INNONE, NULL);
      Delete_IO (IO);
      return;
    }

    if (msg->length > 4096*2) {
      printf("\nInvalid MSG %d\n", msg->length);
      io_set (IO, IO_INNONE, NULL);
      Delete_IO (IO);
      return;
    }

    if (msg->tstamp < irr->start_window) {
      Delete (msg);
      continue;
    }
    if (msg->tstamp > irr->end_window) {
      Delete (msg);
      io_set (IO, IO_INNONE, NULL);
      Delete_IO (IO);
      return;
    }

    switch (msg->type) {
    case MSG_NULL:
      /*print_msg_hdr (msg);*/
      break;

    case MSG_START:
    case MSG_I_AM_DEAD:
    case MSG_PEER_DOWN:
    case MSG_PROTOCOL_IDRP:
    case MSG_PROTOCOL_RIPNG:
    case MSG_PROTOCOL_RIP:
      /*print_msg_hdr (msg);*/
      printf ("*** Not yet implemented ***\n");
      break;

    case MSG_DIE:
      /*print_msg_hdr (msg);*/
      exit(0);
      
    case MSG_PROTOCOL_BGP:
#ifdef HAVE_IPV6
    case MSG_PROTOCOL_BGP4PLUS:
    case MSG_PROTOCOL_BGP4PLUS_01:
#endif /* HAVE_IPV6 */

      if (msg->subtype == MSG_BGP_UPDATE) {
	rtr_process_bgp_update (msg, irr);
	/*print_bgp_msg (msg);*/
      }
      else {
	/*printf ("Something else subtype = %d\n", msg->subtype);*/
      }
      break;

    default:
      printf ("*** Unknown message type %d\n", msg->type);
      break;
    }

    Delete (msg);
  }

  return;
}



/* rtr_process_bgp_update
 * Called from rtr_process_input -- process a individual BGP packet
 *
 */
void rtr_process_bgp_update (mrt_msg_t * msg, irr_connection_t *irr) {
  gateway_t *gateway_to;
  bgp_attr_t *attr;
  LINKED_LIST *ll_with_prefixes, *ll_ann_prefixes;
  prefix_t *prefix;
  short *pref;
  rtracker_peer_t *tmp;
  int new =1;

  attr = NULL;
  ll_with_prefixes = NULL;
  ll_ann_prefixes = NULL;
  pref = NULL;

  bgp_process_update_msg (msg->type, msg->value, msg->length,
			  &gateway_to,
			  &attr,
			  &ll_with_prefixes,
			  &ll_ann_prefixes);


  /* sanity check */
  if (ll_ann_prefixes && (LL_GetCount (ll_ann_prefixes) > 260)) {
    printf ("Too many announced! %d\n", LL_GetCount (ll_ann_prefixes));
    exit (0);
  }
  if (ll_with_prefixes && (LL_GetCount (ll_with_prefixes) > 2000)) {
    printf ("Too many withdrawn %d\n", LL_GetCount (ll_with_prefixes));
    exit (0);
  }


  if ((irr != NULL) && 
      ((irr->output_type == RTR_PREFIXES) || (irr->output_type == RTR_ASCII_PREFIXES))) {
    rtr_run_process (irr, msg->tstamp, attr, ll_with_prefixes, ll_ann_prefixes);
    if (ll_ann_prefixes) 
      LL_Destroy (ll_ann_prefixes);
    if (ll_with_prefixes) 
      LL_Destroy (ll_with_prefixes);
    bgp_deref_attr (attr);
    return;
  }


  /* 
   * otherwise generate a report
   */
  LL_Iterate (irr->ll_peer, tmp) {
    if (prefix_compare (tmp->gateway->prefix, attr->gateway->prefix)) {
      new = 0;
      break;
    }
  }

  if (new) {
    tmp = New (rtracker_peer_t);
    tmp->radix = New_Radix (32);
    tmp->gateway = attr->gateway;
    LL_Add (irr->ll_peer, tmp);
  }

  if (ll_ann_prefixes) {
    /*LL_SetAttributes (ll_ann_prefixes, LL_DestroyFunction, Deref_Prefix, 0);*/
    /* LL_SetAttributes (ll_ann_prefixes, LL_DestroyFunction, Delete_Prefix, 0); */
    LL_Iterate (ll_ann_prefixes, prefix) {
      tmp->total_announce++;
      /*printf ("Announce %s\n", prefix_toax (prefix));*/
      rtr_update_peer (1, tmp, prefix, attr);
    }
    LL_Destroy (ll_ann_prefixes);
  }

  if (ll_with_prefixes) {
    /*LL_SetAttributes (ll_with_prefixes, LL_DestroyFunction, Deref_Prefix, 0);*/
    LL_Iterate (ll_with_prefixes, prefix) {
      tmp->total_withdraw++;
      rtr_update_peer (0, tmp, prefix, attr);
    }
    LL_Destroy (ll_with_prefixes);
  }
  bgp_deref_attr (attr);
  
  return;
}


void rtr_update_peer (int ann_flag, rtracker_peer_t *peer, prefix_t *prefix, bgp_attr_t *attr) {
  radix_node_t *node;
  LINKED_LIST *ll_rtr_attr;
  rtracker_attr_t *rtr_attr, *tmp;
  int new = 1;

  /* just return if processing withdraws */
  if (!ann_flag) {
    return;
  }

  node = radix_lookup (peer->radix, prefix);

  if (node->data == NULL) {
    node->data = (void *) LL_Create (0);
  } 

  ll_rtr_attr = (LINKED_LIST *) node->data;


  LL_Iterate (ll_rtr_attr, rtr_attr) {
    if (bgp_compare_attr (rtr_attr->bgp_attr, attr)) {
      new = 0;
    }
  }

  if (new) {
    tmp = New (rtracker_attr_t);
    tmp->bgp_attr = bgp_ref_attr (attr);
    LL_Add (ll_rtr_attr, tmp);
  }
}


/* report
 */
void report (irr_connection_t *irr) {
  rtracker_peer_t *peer;

  if (irr == NULL) return;

  irr_add_answer (irr, "#%5s %-20s %-8s   %-8s   %-8s   %s\r\n",
		  "AS", "IP", "Unique", "Ann", "With", "Ann+With");
  
  LL_Iterate (irr->ll_peer, peer) {
    irr_add_answer (irr, "%5d %-20s %-8d   %-8d   %-8d   %d\n",
		    peer->gateway->AS, prefix_toa (peer->gateway->prefix), 
		    count_routes (peer),
		    peer->total_announce,  peer->total_withdraw, 
		    peer->total_announce + peer->total_withdraw);
  }
}

int count_routes (rtracker_peer_t *peer) {
  radix_node_t *node;
  int count = 0;

  RADIX_WALK (peer->radix->head, node) {
    if (node->data != NULL) count++;
  }
  RADIX_WALK_END;

  return (count);
}









void rtr_scan_bgp_update (irr_connection_t *irr, mrt_msg_t * msg) {
  gateway_t *gateway_to;
  bgp_attr_t *attr;
  LINKED_LIST *ll_with_prefixes, *ll_ann_prefixes;
  prefix_t *prefix;
  short *pref;
  rtracker_peer_t *tmp;
  int new =1;

  attr = NULL;
  ll_with_prefixes = NULL;
  ll_ann_prefixes = NULL;
  pref = NULL;

  bgp_process_update_msg (msg->type, msg->value, msg->length,
			  &gateway_to,
			  &attr,
			  &ll_with_prefixes,
			  &ll_ann_prefixes);

  LL_Iterate (irr->ll_peer, tmp) {
    if (prefix_compare (tmp->gateway->prefix, attr->gateway->prefix)) {
      new = 0;
      break;
    }
  }

  if (new) {
    tmp = New (rtracker_peer_t);
    tmp->radix = New_Radix (32);
    tmp->gateway = attr->gateway;
    LL_Add (irr->ll_peer, tmp);
  }

  if (ll_ann_prefixes) {
    LL_Iterate (ll_ann_prefixes, prefix) {
      tmp->total_announce++;
      /*printf ("Announce %s\n", prefix_toax (prefix));*/
      rtr_update_peer (1, tmp, prefix, attr);
    }
    Delete (ll_ann_prefixes);
  }

  if (ll_with_prefixes) {
    LL_Iterate (ll_with_prefixes, prefix) {
      tmp->total_withdraw++;
      /*printf ("Withdraw %s\n", prefix_toax (prefix));*/
      rtr_update_peer (0, tmp, prefix, attr);
    }
    Delete (ll_with_prefixes);
  }
}




void rtr_run_process (irr_connection_t *irr, u_long time, 
		      bgp_attr_t *attr, 
		      LINKED_LIST *ll_with_prefixes, 
		      LINKED_LIST *ll_ann_prefixes) {

  char tmp[MAXLINE];
  prefix_t *prefix;
  int ann_flag = 1;
  int with_flag = 1;

  /* check if matching against a given AS */
  if ((irr->AS > 0) && (attr->gateway->AS != irr->AS)) return;

  /* if there are withdrawn prefixes, and we are searching for a specific
   * prefix, the test to see if this list contains that prefix. If it does,
   * set the flag and wee will print this below
   */
  if ((ll_ann_prefixes) && (irr->prefix)) {
    ann_flag = 0;
    LL_Iterate (ll_ann_prefixes, prefix) 
      if (prefix_compare (irr->prefix, prefix)) {
	ann_flag = 1;
	break;
      }
  }
  if ((ll_with_prefixes) && (irr->prefix)) {
    with_flag = 0;
    LL_Iterate (ll_with_prefixes, prefix) 
      if (prefix_compare (irr->prefix, prefix)) {
	with_flag = 1;
	break;
      }
  }
  
  /* announce only */
  if (irr->type == 1) with_flag = 0;
  /* withdraw only */ 
  if (irr->type == 2) ann_flag = 0;


  /* check if we are just searching for an origin AS */
  if ((irr->origin_AS > 0) && (irr->origin_AS != attr->home_AS))
    ann_flag = 0;
  
  strftime (tmp, MAXLINE, "%h %e %T", localtime (&time));

  if (ann_flag && (ll_ann_prefixes)) {
    if (irr->output_type == RTR_ASCII_PREFIXES) {
      irr_add_answer (irr, "%s Announce From AS%d %s\r\n", tmp, attr->gateway->AS, 
		    prefix_toax (attr->gateway->prefix));
      irr_add_answer (irr, "  ASPATH=%s\r\n",  aspath_toa (attr->aspath));
      irr_add_answer (irr, "  NextHop= %s    Origin=%s\r\n", 
		      prefix_toa (attr->nexthop), origin2string (attr->origin));
      if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC)) 
	irr_add_answer (irr, "  MULTIExit=%d\r\n",  attr->multiexit);
      if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY)) 
	irr_add_answer (irr, "  Community=%s\r\n",  attr->community);
       if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR))
	 irr_add_answer (irr, "  Aggregator: AS%d %s\r\n", attr->aggregator.as, 
			 prefix_toa (attr->aggregator.prefix));
    }
    else {
      irr_add_answer (irr, "A|%d|%d|%s|%s\r\n", time, attr->gateway->AS, 
		      prefix_toax (attr->gateway->prefix), bgp_attr_toa (attr));
    }

    LL_Iterate (ll_ann_prefixes, prefix) {
      if ((irr->prefix != NULL) && (!prefix_compare (irr->prefix, prefix))) continue;
      irr_add_answer (irr, "   +%s\r\n", prefix_toax (prefix));
    }
  }

  if (with_flag && (ll_with_prefixes)) {
    if (irr->output_type == RTR_ASCII_PREFIXES) {
      irr_add_answer (irr, "%s Withdraw From AS%d %s\r\n", tmp, attr->gateway->AS, 
		      prefix_toax (attr->gateway->prefix));
    }
    else {
      irr_add_answer (irr, "W|%d|%d|%s\r\n", time, attr->gateway->AS, 
		      prefix_toax (attr->gateway->prefix));
    }
    LL_Iterate (ll_with_prefixes, prefix) {
      if ((irr->prefix != NULL) && (!prefix_compare (irr->prefix, prefix))) continue;
      irr_add_answer (irr, "-%s\r\n", prefix_toax (prefix));
    }
  }

  if (irr->output_type == RTR_ASCII_PREFIXES) 
    irr_add_answer (irr, "\r\n");
}


void Delete_Node (radix_node_t *node) {
  LINKED_LIST *ll_rtr_attr;
  rtracker_attr_t *rtr_attr;

  ll_rtr_attr = (LINKED_LIST *) node->data;

  if (ll_rtr_attr != NULL) {
    LL_SetAttributes (ll_rtr_attr, LL_DestroyFunction, free, 0);
    LL_Iterate (ll_rtr_attr, rtr_attr) {
      bgp_deref_attr (rtr_attr->bgp_attr);
    }
    
    LL_Destroy (ll_rtr_attr);
  }

  if (node->prefix) Deref_Prefix (node->prefix);
  Delete (node);
}

int rtr_delete_radix (rtracker_peer_t *rtr_peer) {
  radix_node_t *node = NULL;
  LINKED_LIST *ll_tmp;

  ll_tmp = LL_Create (LL_DestroyFunction, Delete_Node, 0);

  if (rtr_peer->radix != NULL) {
    RADIX_WALK_ALL (rtr_peer->radix->head, node) {
      LL_Add (ll_tmp, node);
    }
    RADIX_WALK_END;
  }
  else
    printf ("SHould never get here\n");


  /*Aprintf ("%s %d\nra", prefix_toa (rtr_peer->gateway->prefix), LL_GetCount (ll_tmp));*/

  LL_Destroy (ll_tmp);
  if (rtr_peer->radix != NULL) {
    Delete (rtr_peer->radix);
  }
  return (1);
}


void Delete_Peer (rtracker_peer_t *peer) {
  
  rtr_delete_radix (peer);
  Delete (peer);
}
